// Copyright 2022 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// All of the actual test files have limited build constraints. This file
// ensures there's at least one test file on every platform.

package sanitizers_test
