// Copyright 2022 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// This file is here to silence an error about registerCache not having a body.
// (The body is provided by package runtime.)
