// Code generated by cmd/cgo -godefs; DO NOT EDIT.
// cgo -godefs defs_darwin.go

package route

const (
	sizeofIfMsghdrDarwin15    = 0x70
	sizeofIfaMsghdrDarwin15   = 0x14
	sizeofIfmaMsghdrDarwin15  = 0x10
	sizeofIfMsghdr2Darwin15   = 0xa0
	sizeofIfmaMsghdr2Darwin15 = 0x14
	sizeofIfDataDarwin15      = 0x60
	sizeofIfData64Darwin15    = 0x80

	sizeofRtMsghdrDarwin15  = 0x5c
	sizeofRtMsghdr2Darwin15 = 0x5c
	sizeofRtMetricsDarwin15 = 0x38

	sizeofSockaddrStorage = 0x80
	sizeofSockaddrInet    = 0x10
	sizeofSockaddrInet6   = 0x1c
)
