// Code generated by mkduff.go; DO NOT EDIT.
// Run go generate from src/runtime to update.
// See mkduff.go for comments.

#include "textflag.h"

TEXT runtime·duffzero(SB), NOSPLIT, $0-0
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	MOVW.P	R0, 4(R1)
	RET

TEXT runtime·duffcopy(SB), NOSPLIT, $0-0
	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	MOVW.P	4(R1), R0
	MOVW.P	R0, 4(R2)

	RET
