// Copyright 2016 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package goarch

const (
	_ArchFamily          = S390X
	_DefaultPhysPageSize = 4096
	_PCQuantum           = 2
	_MinFrameSize        = 8
	_StackAlign          = PtrSize
)
