// Copyright 2020 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// This assembly file exists to allow internal/cpu to call
// non-exported runtime functions that use "go:linkname".