// Copyright 2023 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// This file exists only to drive go:generate.
//go:generate go test -run=Generate -write=all

package types
